package PimmelServ;

use strict;
use Data::Dumper;

my %clients = (
	PimmelServ => {
		nick			=> 'PimmelServ',
		ident			=> 'pimmel',
		host			=> 'serv',
		ip				=> '0.0.0.0',
		cloak			=> 'serv',
		gecos			=> 'Pimmel Service',
		umode			=> '+Siopy',
		authflags	=> '$=^_><|',
		nickserv		=> 'password',
		chans			=> {
								'#acid'	=> ''
							}
	}
);

my ($config, %chans, %pending) = 'mod/PimmelServ.conf';

sub new {
	my ($package, %self) = @_;
	my $self = bless(\%self, $package);

	acid::setupclients(%clients);

	my $uid = acid::getid($clients{PimmelServ}{nick});

	if (my $error = chans_load()) {
		acid::privmsg($uid, $self->{conf}{logchan}, 'Failed to load %s: %s', $config, $error);
	}
	else {
		acid::privmsg($uid, $self->{conf}{logchan}, 'Joining %d channel entries from %s.', scalar(keys(%chans)), $config);
		acid::join_($uid, $_, '+nt', '') for keys(%chans);
	}

	return $self;
}

sub on_unload {
	my $self = shift;

	acid::removeclients(%clients);
}

sub chans_load {
	my %test = do($config);

	return $! if $!;

	if ($@) {
		chomp $@;
		return "$@";
	}

	%chans = %test;

	return;
}

sub chans_save {
	local $Data::Dumper::Sortkeys = 1;
	my @dump = split(/\n/, Dumper(\%chans));

	open CHANS, '>', $config || return $!;
	print CHANS substr($_, 10)."\n" for splice(@dump, 1, -1);
	close CHANS;

	return;
}

sub on_privmsg {
	my ($self, $uid, $target, $msg) = @_;

	my $targetnick = acid::getname($target) if substr($target, 0, 1) ne '#';
	my $nick = acid::getname($uid);

	if ($targetnick && $targetnick eq $clients{PimmelServ}{nick}) {
		my @argv = split(' ', $msg);
		my $cmd = uc(shift(@argv));

		if ($cmd eq 'HELP') {
			acid::notice($target, $uid, qq(
\002$targetnick\002 allows you to have yourself pimmelt an.
 
To request \002$targetnick\002 for a channel, type
\002/msg $targetnick REQUEST \037channel\037\002
 
To remove \002$targetnick\002 from a channel, type
\002/msg $targetnick REMOVE \037channel\037\002
 
\002Note:\002 The above commands require you to have founder
access to the given channel.
));
		}
		elsif ($cmd eq 'REQUEST' || $cmd eq 'REMOVE') {
			my $type = $cmd eq 'REQUEST' ? 1 : 0;

			if (!@argv || substr($argv[0], 0, 1) ne '#') {
				acid::notice($target, $uid, "Syntax: \002%s \037channel\037\002", $cmd);
				return;
			}

			my $chan = $self->{lc}{lc($argv[0])};

			if (!$chan) {
				acid::notice($target, $uid, "Channel \002%s\002 does not exist.", $argv[0]);
				return;
			}

			if ($type && $chans{lc($chan)}) {
				acid::notice($target, $uid, "I already am on channel \002%s\002.", $argv[0]);
				return;
			}

			if (!$type && !$chans{lc($chan)}) {
				acid::notice($target, $uid, "I am not on channel \002%s\002.", $argv[0]);
				return;
			}

			if ($pending{$chan}{$nick}) {
				acid::notice($target, $uid, "There already is a pending request from you for channel \002%s\002 from \002%s\002 ago.", $chan, acid::duration(time - $pending{$chan}{$nick}{ts}));
				return;
			}

			$pending{$chan}{$nick}{ts} = time;
			$pending{$chan}{$nick}{type} = $type;
			acid::notice($target, $uid, "Checking if you have founder access to channel \002%s\002.", $argv[0]);
			acid::privmsg($target, $self->{conf}{logchan}, 'Received %s from user %s for channel %s.', $cmd, $nick, $chan);
			acid::privmsg($target, 'ChanServ@services.rizon.net', 'STATUS %s %s', lc($chan), $nick);
		}
	}
	elsif ($chans{$target} && substr($msg, 0, 1) eq '.') {
		my $myuid = acid::getid($clients{PimmelServ}{nick});

		my @args = split(' ', $msg);
		my $cmd = uc(substr(shift(@args), 1));

		if ($cmd eq 'PIMMEL') {
			acid::privmsg($myuid, $target, "\001ACTION pimmelt \002\00313%s\003\002 an.\001", $nick);
		}
	}
}

sub on_notice {
	my ($self, $uid, $target, $msg) = @_;

	my $targetnick = acid::getname($target) if substr($target, 0, 1) ne '#';
	my $nick = acid::getname($uid);

	if ($targetnick eq $clients{PimmelServ}{nick} && $nick eq 'ChanServ') {
		my @argv = split(' ', $msg);

		if ($msg eq 'Access denied.') {
			acid::privmsg($target, $self->{conf}{logchan}, 'Received ACCESS DENIED from %s.', $nick);
		}
		elsif ($argv[0] eq 'STATUS') {
			if ($argv[1] eq 'ERROR') {
				acid::privmsg($target, $self->{conf}{logchan}, 'Received STATUS ERROR from %s: %s', $nick, "@argv[2..$#argv]");
			}
			elsif (!$pending{$argv[1]}{$argv[2]}) {
				acid::privmsg($target, $self->{conf}{logchan}, 'Received STATUS reply for non-pending nick/channel from %s: %s', $nick, $msg);
			}
			elsif ($argv[3] < 10000) {
				acid::notice($target, acid::getid($argv[2]), "Request denied for channel \002%s\002 due to insufficient access, requesting %s requires founder access.", $argv[1], $targetnick);
				acid::privmsg($target, $self->{conf}{logchan}, 'Request from user %s for channel %s denied: access %d is insufficient', $argv[2], $argv[1], $argv[3]);
				delete $pending{$argv[1]}{$argv[2]};
			}
			else {
				if ($pending{$argv[1]}{$argv[2]}{type}) {
					acid::notice($target, acid::getid($argv[2]), "Request accepted for channel \002%s\002, see you in there!", $argv[1]);
					acid::privmsg($target, $self->{conf}{logchan}, 'Request from user %s for channel %s accepted, access %d is sufficient.', $argv[2], $argv[1], $argv[3]);
					acid::join_($target, $argv[1], '+nt', '');

					$chans{$argv[1]}{nick} = $argv[2];
					$chans{$argv[1]}{ts} = $pending{$argv[1]}{$argv[2]}{ts};
				}
				else {
					acid::notice($target, acid::getid($argv[2]), "Remove request accepted for channel \002%s\002, bye bye!", $argv[1]);
					acid::privmsg($target, $self->{conf}{logchan}, 'Remove request from user %s for channel %s accepted, access %d is sufficient.', $argv[2], $argv[1], $argv[3]);
					acid::part($target, $argv[1], "REMOVE from $argv[2]");

					delete $chans{$argv[1]};
				}

				delete $pending{$argv[1]}{$argv[2]};

				if (my $error = chans_save()) {
					acid::privmsg($uid, $self->{conf}{logchan}, 'Failed to save %s: %s', $config, $error);
				}
			}
		}
	}
}

1;
