package RootServ;

use strict;
use Data::Dumper;

my %clients = (
	RootServ	=> {
		nick			=> 'RootServ',
		ident			=> 'root',
		host			=> 'acid.rizon.net',
		ip				=> '0.0.0.0',
		cloak			=> 'acid.rizon.net',
		gecos			=> 'RootServ',
		umode			=> '+NSaioqy',
		authflags	=> '$=^_><|',
		nickserv		=> 'password',
		chans			=> {
								'#acid'	=> '',
								'#opers'	=> '@',
								'#r'		=> ''
							}
	}
);

my %flag = (
	AUTHFLAGS	=> 'a',

	CHECKBAN		=> 'b',

	CONFIG		=> 'c',

	SHUTDOWN		=> 'd',

	DUMP			=> 'e',
	EVAL			=> 'e',
	MODULE		=> 'e',

	FLAGS			=> 'f',

	CHANLIST		=> 'i',
	FIND			=> 'i',
	GETINFO		=> 'i',
	USERLIST		=> 'i',

	KILL			=> 'k',

	KICK			=> 'm',
	MODE			=> 'm',

	CAPTURE		=> 'p',
	UNCAPTURE	=> 'p',

	RUSERLIST	=> 'r',

	SVSJOIN		=> 's',
	SVSMODE		=> 's',
	SVSNICK		=> 's',
	SVSPART		=> 's',

	RAW			=> 'w'
);

my %help;
for (<help/RootServ/*>) {
	my ($cmd, @help) = substr($_, 14);
	open HELP, '<', $_;
	for (<HELP>) {
		chomp;
		s/%SERV%/$clients{RootServ}{nick}/g;
		push @help, $_;
	}
	close HELP;
	$help{$cmd}{syntax} = shift(@help);
	$help{$cmd}{desc} = shift(@help);
	$help{$cmd}{help} = join("\n", @help);
}

sub new {
	my ($package, %self) = @_;
	my $self = bless(\%self, $package);

	acid::setupclients(%clients);
	acid::privmsg(acid::getid($clients{RootServ}{nick}), $self->{conf}{logchan}, 'Finished bursting after %s.', acid::duration(time - $^T));

	return $self;
}

sub getflags {
	my ($self, $target) = @_;

	for (keys(%{$self->{conf}{flags}})) {
			if (lc($_) eq lc($self->{client}{$target}{su}) || $_ eq $self->{client}{$target}{certfp} || lc($_) eq lc($target)) {
				return $self->{conf}{flags}{$_}
			}
	}

	return;
}

sub hasaccess {
	my ($self, $cmd, $uid) = @_;

	return 1 if $help{$cmd} && !$flag{$cmd};

	my $flags = $self->getflags($uid);

	return 1 if $flags && $flags =~ /$flag{$cmd}/;

	return;
}

sub on_unload {
	my $self = shift;

	acid::removeclients(%clients);
}

sub on_whois {
	my ($self, $sid, $uid, $target) = @_;

	acid::privmsg($target, $self->{conf}{logchan}, '%s is doing a WHOIS on me via %s.', acid::getname($uid), acid::getname($sid));
}

sub on_stats {
	my ($self, $id, $letter, $target) = @_;

	acid::privmsg(acid::getid($clients{RootServ}{nick}), $self->{conf}{logchan}, '%s is doing a STATS %s on %s.', acid::getname($id), $letter, $target);

	if (uc($letter) eq 'O') {
		for (sort(keys(%{$self->{conf}{flags}}))) {
			acid::raw($self->{conf}{sid}, '243 %s %s :* * %s %s 0', $id, uc($letter), $_, $self->{conf}{flags}{$_});
		}
		acid::raw($self->{conf}{sid}, '219 %s %s :End of /STATS report', $id, $letter);
	}
}

sub on_version {
	my ($self, $id, $target) = @_;

	acid::privmsg(acid::getid($clients{RootServ}{nick}), $self->{conf}{logchan}, '%s is doing a VERSION on %s.', acid::getname($id), $target);
}

sub on_time {
	my ($self, $id, $target) = @_;

	acid::privmsg(acid::getid($clients{RootServ}{nick}), $self->{conf}{logchan}, '%s is doing a TIME on %s.', acid::getname($id), $target);
}

sub on_ctcp {
	my ($self, $uid, $target, $ctcp) = @_;

	if (substr($target, 0, 1) ne '#') {
		acid::privmsg($target, $self->{conf}{logchan}, '%s is doing a CTCP %s on me.', acid::getname($uid), $ctcp);
	}
}

sub on_privmsg {
	my ($self, $uid, $target, $msg) = @_;

	if (substr($target, 0, 1) ne '#') {

		my $targetnick = acid::getname($target) if substr($target, 0, 1) ne '#';
		my $nick = acid::getname($uid);

		$msg =~ s/([\xC0-\xDF])([\x80-\xBF])/chr(ord($1)<<6&0xC0|ord($2)&0x3F)/eg;

		acid::privmsg($target, $self->{conf}{logchan}, '%s: %s', $nick, $msg);

		if ($targetnick && $targetnick eq $clients{RootServ}{nick}) {
			if (!$self->{client}{$uid}{umode}{o}) {
				acid::operwall($target, "Access denied to non-oper \002%s\002 (%s@%s).", $nick, $self->{client}{$uid}{ident}, $self->{client}{$uid}{host});
				return;
			}

			my @argv = split(' ', $msg);
			my $cmd = uc(shift(@argv));

			if (!$help{$cmd}) {
				acid::notice($target, $uid,  "\002%s\002 is not a known command, see \002/msg %s HELP\002.", $cmd, $targetnick);
				return;
			}

			if (!$self->hasaccess($cmd, $uid)) {
				acid::notice($target, $uid, "Access denied. Command \002%s\002 requires oper flag \002%s\002.", $cmd, $flag{$cmd});
				return;
			}

			# HELP
			if ($cmd eq 'HELP') {
				if (!@argv) {
					my $flags = $self->getflags($uid);

					acid::notice($target, $uid, "%s\n ", $help{$cmd}{help});

					for (sort { $flag{$a} cmp $flag{$b} || $a cmp $b } keys(%help)) {
						acid::notice($target, $uid, '  [%s] %-10s %s', ($flag{$_} ? $flag{$_} : ' '), $_, $help{$_}{desc});
					}

					if ($flags) {
						acid::notice($target, $uid, " \nYour flags are \002%s\002.", $flags);
					}
					else {
						acid::notice($target, $uid, " \nYou don't have any flags. At all. Not a single one.");
					}

					return;
				}

				my $help = uc($argv[0]);
				if (!$help{$help}) {
					acid::notice($target, $uid, "No help available for \002%s\002.", $help);
					return;
				}

				acid::notice($target, $uid, "Syntax: %s\n \n%s\n ", $help{$help}{syntax}, $help{$help}{help});

				if ($flag{$help}) {
					acid::notice($target, $uid, "This command requires oper flag \002%s\002.", $flag{$help});
				}
				else {
					acid::notice($target, $uid, 'This command does not require any oper flag.');
				}
			}

			# DUMP
			elsif ($cmd eq 'DUMP') {
				if (!@argv) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				local $Data::Dumper::Sortkeys = 1;

				if ($self->{conf}{livenet}) {
					my $dump = time.'.dump';
					unless (open DUMP, '>', $dump) {
						acid::notice($target, $uid, "Failed to dump \002%s\002: %s", $argv[0], $!);
						return;
					}
					printf DUMP "Dump of %s requested by %s at %s:\n\n%s", $argv[0], $nick, scalar(localtime), Dumper(eval($argv[0]));
					close DUMP;
					acid::notice($target, $uid, "\002%s\002 has been dumped to \002%s\002.", $argv[0], $dump);
				}
				else {
					acid::notice($target, $uid, '%s', Dumper(eval($argv[0])));
				}
			}

			# EVAL
			elsif ($cmd eq 'EVAL') {
				if (!@argv) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				if ($self->{conf}{livenet}) {
					acid::notice($target, $uid, "Command \002%s\002 may only be used in debug mode.", $cmd);
					return;
				}

				eval "@argv";
				return if !$@;
				chomp $@;
				acid::notice($target, $uid, "Compilation of %s \002%s\002 has failed:\n%s", $cmd, "@argv", $@);
			}

			# RAW
			elsif ($cmd eq 'RAW') {
				if (!@argv) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				if ($self->{conf}{livenet}) {
					acid::notice($target, $uid, "Command \002%s\002 may only be used in debug mode.", $cmd);
					return;
				}

				acid::raw(undef, '%s', "@argv");
				acid::notice($target, $uid, "Raw command \002%s\002 has been sent.", "@argv");
				acid::operwall($target, "\002%s\002 used \002%s\002.", $nick, $cmd) if !$self->{conf}{abuse};
			}

			# STATS
			elsif ($cmd eq 'STATS') {
				my $startup = localtime($^T);
				$startup =~ s/ {2,}/ /g;

				my $load = 'n/a';
				if ($^O eq 'linux') {
					open LOAD, '<', '/proc/loadavg';
					$load = join(' ', (split(/ /, <LOAD>))[0..2]);
					close LOAD;
				}
				elsif ($^O eq 'freebsd') {
					$load = join(' ', (split(/ /, `sysctl vm.loadavg`))[2..4]);
				}

				my $opers = 0;
				for (keys(%{$self->{client}})) {
					$opers++ if $self->{client}{$_}{umode}{o};
				}

				acid::notice($target, $uid, qq(
\002Stats for %s (%s):\002
 
\002Version:\002  acidictive-%s
\002Uplink:\002   %s (%s)
\002Uptime:\002   %s (%s)
\002Load:\002     %s
\002Modules:\002  %d (%s)
\002Logchan:\002  %s (+%s)
\002Debuglog:\002 %s (%d KB)
 
\002Servers:\002  %d
\002Channels:\002 %d
\002Users:\002    %d (%d opers)
), $self->{conf}{name},
	$self->{conf}{sid},
	$acid::VERSION,
	$self->{server}{$acid::uplink},
	$acid::uplink, acid::duration(time - $^T),
	$startup,
	$load,
	scalar(keys(%{$self->{module}})),
	join(', ', sort(keys(%{$self->{module}}))),
	$self->{conf}{logchan},
	join('', sort(keys(%{$self->{chan}{$self->{lc}{lc($self->{conf}{logchan})}}{mode}}))),
	($self->{conf}{debuglog} ? $self->{conf}{debuglog} : 'disabled'),
	($self->{conf}{debuglog} ? int((-s $self->{conf}{debuglog}) / 1024) : 0),
	scalar(keys(%{$self->{server}})),
	scalar(keys(%{$self->{chan}})),
	scalar(keys(%{$self->{client}})),
	$opers
				);
			}

			# KILL
			elsif ($cmd eq 'KILL') {
				if (!@argv) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my $victim = acid::getid($argv[0]);
				if (!$victim) {
					acid::notice($target, $uid, "There is no user \002%s\002 connected.", $argv[0]);
					return;
				}

				if ($self->{client}{$victim}{sid} eq $self->{conf}{sid}) {
					acid::notice($target, $uid, "Why would you want to kill \002%s\002?", $argv[0]);
					return;
				}

				#if ($self->{client}{$victim}{umode}{N} && $victim ne $uid) {
				#	acid::notice($target, $uid, "Command \002%s\002 may not be used on other network administrators.", $cmd);
				#	return;
				#}

				acid::kill($victim, "@argv[1..$#argv]");
				acid::notice($target, $uid, "User \002%s\002 has been killed.", $argv[0]);
				acid::operwall($target, "\002%s\002 used \002%s\002 on \002%s\002.", $nick, $cmd, $argv[0]) if !$self->{conf}{abuse};
			}

			# SHUTDOWN
			elsif ($cmd eq 'SHUTDOWN') {
				acid::shutdown_($target, $nick);
			}

			# CONFIG
			elsif ($cmd eq 'CONFIG') {
				if (uc($argv[0]) eq 'LIST') {
					if ($argv[1]) {
						if (defined($self->{conf}{$argv[1]})) {
							acid::notice($target, $uid, "Config option \002%s\002 is set to \002%s\002.", $argv[1], $self->{conf}{$argv[1]});
						}
						else {
							acid::notice($target, $uid, "Option \002%s\002 could not be found in the config.", $argv[1]);
						}
					}
					else {
						local $Data::Dumper::Sortkeys = 1;
						my @dump = split(/\n/, Dumper($self->{conf}));
						acid::notice($target, $uid, '%s', substr($_, 10)) for splice(@dump, 1, -1);
					}
				}
				elsif (uc($argv[0]) eq 'SET') {
					if (!defined($argv[2])) {
						acid::notice($target, $uid, "Syntax: \002%s %s \037option\037 \037value\037\002", $cmd, $argv[0]);
						return;
					}

					my $value = "@argv[2..$#argv]";
					$self->{conf}{$argv[1]} = $value;
					acid::notice($target, $uid, "Config option \002%s\002 has been set to \002%s\002.", $argv[1], $value);
				}
				elsif (uc($argv[0]) eq 'UNSET') {
					if (!$argv[1]) {
						acid::notice($target, $uid, "Syntax: \002%s %s \037option\037\002", $cmd, $argv[0]);
						return;
					}

					if (!defined($self->{conf}{$argv[1]})) {
						acid::notice($target, $uid, "Option \002%s\002 is not defined in the config.", $argv[1]);
						return;
					}

					delete $self->{conf}{$argv[1]};
					acid::notice($target, $uid, "Config option \002%s\002 has been unset.", $argv[1]);
				}
				elsif (uc($argv[0]) eq 'SAVE') {
					my $error = acid::config_save();
					if ($error) {
						acid::notice($target, $uid, 'Config could not be saved: %s', $error);
					}
					else {
						acid::notice($target, $uid, 'Config has been saved.');
					}
				}
				elsif (uc($argv[0]) eq 'REHASH') {
					my $error = acid::config_load();
					if ($error) {
						acid::notice($target, $uid, 'Config could not be rehashed: %s', $error);
					}
					else {
						acid::notice($target, $uid, 'Config has been rehashed.');
					}
				}
				else {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
				}
			}

			# MODULE
			elsif ($cmd eq 'MODULE') {
				if (uc($argv[0]) eq 'LOAD') {
					if (!$argv[1]) {
						acid::notice($target, $uid, "Syntax: \002%s %s \037module\037\002", $cmd, $argv[0]);
						return;
					}

					if ($self->{module}{$argv[1]}) {
						acid::notice($target, $uid, "Module \002%s\002 is already loaded.", $argv[1]);
						return;
					}

					my $error = acid::module_load($argv[1]);
					if ($error) {
						acid::notice($target, $uid, "Module \002%s\002 could not be loaded:\n%s", $argv[1], $error);
					}
					else {
						acid::notice($target, $uid, "Module \002%s\002 has been loaded.", $argv[1]);
					}
				}
				elsif (uc($argv[0]) eq 'UNLOAD') {
					if (!$argv[1]) {
						acid::notice($target, $uid, "Syntax: \002%s %s \037module\037\002", $cmd, $argv[0]);
						return;
					}

					if (!$self->{module}{$argv[1]}) {
						acid::notice($target, $uid, "There is no \002%s\002 module loaded.", $argv[1]);
						return;
					}

					acid::module_prepare_unload($argv[1]);
					acid::notice($target, $uid, "Module \002%s\002 has been unloaded.", $argv[1]);
				}
				elsif (uc($argv[0]) eq 'REFRESH') {
					if (!$argv[1]) {
						acid::notice($target, $uid, "Syntax: \002%s %s \037module\037\002", $cmd, $argv[0]);
						return;
					}
					my $error = acid::module_refresh($argv[1]);
					if ($error) { 
						acid::notice($target, $uid, "Module \002%s\002 could not be refreshed:\n%s", $argv[1], $error);
					}
					else {
						acid::notice($target, $uid, "Module \002%s\002 has been refreshed.", $argv[1]);
					}
				}
				elsif (uc($argv[0]) eq 'LIST') {
					my $count;
					for (sort { lc($a) cmp lc($b) } <mod/*.pm>) {
						my ($module, @stat) = (substr($_, 4, -3), stat);
						acid::notice($target, $uid, '%s %s (%d KB, created on %s, last modified %s ago)',
							($self->{module}{$module} ? '*' : ' '), $module, int($stat[7] / 1024), scalar(localtime($stat[8])), acid::duration(time - $stat[9]));
						$count++;
					}
					acid::notice($target, $uid, "Listed \002%d\002 modules.", $count);
				}
				else {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
				}
			}

			# FLAGS
			elsif ($cmd eq 'FLAGS') {
				if ($argv[1]) {
					if ($argv[1] eq '-') {
						for (keys(%{$self->{conf}{flags}})) {
							if (lc($_) eq lc($argv[0])) {
								delete $self->{conf}{flags}{$_};
								acid::notice($target, $uid, "Flags of \002%s\002 have been removed.", $argv[0]);
								return;
							}
						}

						acid::notice($target, $uid, "User \002%s\002 doesn't have any flags.", $argv[0]);
					}
					else {
						(my $flags = join('', sort(split(//, $argv[1])))) =~ s/(.)(?=.*?\1)//g;
						(my $avail = join('', sort(values(%flag)))) =~ s/(.)(?=.*?\1)//g;
						$flags = $avail if $flags =~ /\+/;
						if ($flags !~ /^[+$avail]+$/) {
							acid::notice($target, $uid, "\002%s\002 is not a valid set of flags, please see \002/msg %s HELP\002", $argv[1], $target);
							return;
						}

						for (keys(%{$self->{conf}{flags}})) {
							if (lc($_) eq lc($argv[0])) {
								delete $self->{conf}{flags}{$_};
								last;
							}
						}

						$self->{conf}{flags}{$argv[0]} = $flags;
						acid::notice($target, $uid, "Flags of user \002%s\002 have been set to \002%s\002.", $argv[0], $flags);
					}
				}
				elsif ($argv[0]) {
					if (my $flags = $self->getflags($argv[0])) {
						acid::notice($target, $uid, "Flags of user \002%s\002 are \002%s\002.", $argv[0], $flags);
					}
					else {
						acid::notice($target, $uid, "User \002%s\002 doesn't have any flags.", $argv[0]);
					}
				}
				else {
					my ($count, @certfps);
					for (sort { lc($a) cmp lc($b) } keys(%{$self->{conf}{flags}})) {
						if (length($_) == 40) {
							push @certfps, $_;
							next;
						}

						$count++;
						acid::notice($target, $uid, '%02d. %s - %s', $count, $_, $self->{conf}{flags}{$_});
					}

					for (@certfps) {
						$count++;
						acid::notice($target, $uid, '%02d. %s - %s (CertFP)', $count, $_, $self->{conf}{flags}{$_});
					}

					acid::notice($target, $uid, "Listed the flags of \002%d\002 opers.", $count);
				}
			}

			# USERLIST
			elsif ($cmd eq 'USERLIST') {
				if (!@argv || "@argv" !~ /.!.+@./) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my %ulist;

				$ulist{mask} = quotemeta("@argv");
				$ulist{mask} =~ s/\\\*/.*/g;
				$ulist{mask} =~ s/\\\?/./g;

				($ulist{nick}, $ulist{ident}, $ulist{host}, $ulist{gecos}) = split(/\\[:@!]/, $ulist{mask}, 4);
				$ulist{gecos} ||= '.*';

				$ulist{count} = 0;

				for (sort { lc($self->{client}{$a}{nick}) cmp lc($self->{client}{$b}{nick}) } keys(%{$self->{client}})) {
					if ($self->{client}{$_}{nick} =~ /^$ulist{nick}$/i
					&& $self->{client}{$_}{ident} =~ /^$ulist{ident}$/i
					&&	$self->{client}{$_}{gecos} =~ /^$ulist{gecos}$/i
					&&	($self->{client}{$_}{host} =~ /^$ulist{host}$/i
						|| $self->{client}{$_}{ip} =~ /^$ulist{host}$/i
						|| $self->{client}{$_}{cloak} =~ /^$ulist{host}$/i)
					) {
						$ulist{count}++;
						acid::notice($target, $uid, '%s', acid::getinfo($_));
					}
				}

				acid::notice($target, $uid, "Listed \002%d\002/\002%d\002 users matching \002%s\002.",
					$ulist{count}, scalar(keys(%{$self->{client}})), "@argv");
			}

			# RUSERLIST
			elsif ($cmd eq 'RUSERLIST') {
				if (!@argv || "@argv" !~ /.!.+@.+:./) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my %rulist;

				$rulist{mask} = "@argv";
				eval { qr($rulist{mask}) };
				if ($@) {
					chomp $@;
					acid::notice($target, $uid, "Please have your \002%s\002 flag removed \002immediately\002:\n%s", $flag{$cmd}, $@);
					acid::operwall($target, "\002%s\002 miserably fails at regular expressions and needs their \002%s\002 flag removed.", $nick, $flag{$cmd});
					return;
				}

				$rulist{count} = 0;

				for (sort { lc($self->{client}{$a}{nick}) cmp lc($self->{client}{$b}{nick}) } keys(%{$self->{client}})) {
					$rulist{cloak} = "$self->{client}{$_}{nick}!$self->{client}{$_}{ident}\@$self->{client}{$_}{cloak}:$self->{client}{$_}{gecos}";
					$rulist{host} = "$self->{client}{$_}{nick}!$self->{client}{$_}{ident}\@$self->{client}{$_}{host}:$self->{client}{$_}{gecos}";
					$rulist{ip} = "$self->{client}{$_}{nick}!$self->{client}{$_}{ident}\@$self->{client}{$_}{ip}:$self->{client}{$_}{gecos}";

					if ($rulist{cloak} =~ /^$rulist{mask}$/ || $rulist{host} =~ /^$rulist{mask}$/ || $rulist{ip} =~ /^$rulist{mask}$/) {
						$rulist{count}++;
						acid::notice($target, $uid, '%s', acid::getinfo($_));
					}
				}
				acid::notice($target, $uid, "Listed \002%d\002/\002%d\002 users matching \002%s\002.", $rulist{count}, scalar(keys(%{$self->{client}})), "@argv");
			}

			# CHANLIST
			elsif ($cmd eq 'CHANLIST') {
				if (!@argv || substr($argv[0], 0, 1) ne '#') {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				if (!$self->{lc}{lc($argv[0])}) {
					acid::notice($target, $uid, "Channel \002%s\002 does not exist.", $argv[0]);
					return;
				}

				my %chan = %{$self->{chan}{$self->{lc}{lc($argv[0])}}};

				for (sort { $self->{prefix}{substr($chan{users}{$a}, 0, 1)} <=> $self->{prefix}{substr($chan{users}{$b}, 0, 1)}
					|| lc($self->{client}{$a}{nick}) cmp lc($self->{client}{$b}{nick}) } keys(%{$chan{users}}))
				{
					$chan{count}++;
					if ($self->{client}{$_}{umode}{o}) {
						$chan{opers}++;
						acid::notice($target, $uid, "\002%s%s\002", $chan{users}{$_}, acid::getinfo($_));
					}
					else {
						acid::notice($target, $uid, '%s%s', $chan{users}{$_}, acid::getinfo($_));
					}
				}
				$chan{modes} = join('', sort(keys(%{$chan{mode}})));
				$chan{modes} .= " $chan{mode}{k}" if $chan{mode}{k};
				$chan{modes} .= " $chan{mode}{l}" if $chan{mode}{l};
				if ($chan{count}) {
					acid::notice($target, $uid, "Listed \002%d\002 users%s currently being on \002%s\002 (+%s).",
						$chan{count}, ($chan{opers} ? " including \002$chan{opers}\002 opers" : ''), $argv[0], $chan{modes});
				}
				else {
					acid::notice($target, $uid, "Channel \002%s\002 (+%s) is currently empty.", $argv[0], $chan{modes});
				}
			}

			# GETINFO
			elsif ($cmd eq 'GETINFO') {
				if (!@argv) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my $infouid = acid::getid($argv[0]);

				if (!$infouid) {
					acid::notice($target, $uid, "There is no user \002%s\002 connected.", $argv[0]);
					return;
				}

				my $first = 1;

				for my $u ($infouid, sort { lc($self->{client}{$a}{nick}) cmp lc($self->{client}{$b}{nick}) } keys(%{$self->{client}})) {
					next if $u eq $infouid && !$first;
					$first = 0 if $first;
					if ($self->{client}{$u}{host} eq $self->{client}{$infouid}{host}
						|| ($self->{client}{$u}{ip} eq $self->{client}{$infouid}{ip} && $self->{client}{$u}{ip} && $self->{client}{$u}{ip} ne '255.255.255.255')) {
						acid::notice($target, $uid, '%s%s / %s / su: %s / authflags: %s',
															($u ne $infouid ? " \nClone: " : ''),
															acid::getinfo($u),
															($self->{client}{$u}{prevnicks} ? join(' -> ', @{$self->{client}{$u}{prevnicks}}, $self->{client}{$u}{nick}) : 'no nick changes'),
															($self->{client}{$u}{su} ? $self->{client}{$u}{su} : 'none'),
															($self->{client}{$u}{authflags} ? $self->{client}{$u}{authflags} : 'none')
						);
						my $chans;
						$chans .= ($self->{chan}{$_}{mode}{s} ? '!' : '')."$self->{client}{$u}{chans}{$_}$_ " for sort { lc($a) cmp lc($b) } keys(%{$self->{client}{$u}{chans}});
						$chans = ($chans ? substr($chans, 0, -1) : 'none');
						acid::notice($target, $uid, 'Channels: %s', $chans) if $chans;
						acid::notice($target, $uid, 'CertFP: %s', $self->{client}{$u}{certfp}) if $self->{client}{$u}{certfp};
					}
				}
			}

			# FIND
			elsif ($cmd eq 'FIND') {
				if (!@argv) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my $count;
				for my $u (keys(%{$self->{client}})) {
					if (lc($self->{client}{$u}{nick}) eq lc($argv[0])) {
						$count++;
						acid::notice($target, $uid, 'Online: %s', acid::getinfo($u));
						next;
					}
					next if !$self->{client}{$u}{prevnicks};
					for (@{$self->{client}{$u}{prevnicks}}) {
						if (lc($_) eq lc($argv[0]) && lc($_) ne lc($self->{client}{$u}{nick})) {
							$count++;
							acid::notice($target, $uid, '%s / %s', acid::getinfo($u), join(' -> ', @{$self->{client}{$u}{prevnicks}}, $self->{client}{$u}{nick}));
							last;
						}
					}
				}

				if ($count) {
					acid::notice($target, $uid, "Listed \002%d\002 users being or having previously been known as \002%s\002.", $count, $argv[0]);
				}
				else {
					acid::notice($target, $uid, "Nick \002%s\002 hasn't been used by anyone who currently is connected.", $argv[0]);
				}

			}

			# CHECKBAN
			elsif ($cmd eq 'CHECKBAN') {
				if (!@argv) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my %target;

				if ($argv[0] =~ /.!.+@./) {
					$target{info} = $argv[0];
					($target{nick}, $target{ident}, $target{host}) = split(/[!@]/, $argv[0]);
				}
				elsif ($target{uid} = acid::getid($argv[0])) {
					$target{info} = acid::getinfo($target{uid});
					$target{nick} = $self->{client}{$target{uid}}{nick};
					$target{ident} = $self->{client}{$target{uid}}{ident};
					$target{cloak} = $self->{client}{$target{uid}}{cloak};
					$target{host} = $self->{client}{$target{uid}}{host};
					$target{ip} = $self->{client}{$target{uid}}{ip};
				}
				else {
					acid::notice($target, $uid, "There is no user \002%s\002 connected.", $argv[0]);
					return;
				}

				@{$target{chans}} = $argv[1] ? @argv[1..$#argv] : sort { lc($a) cmp lc($b) } keys(%{$self->{chan}});
				acid::notice($target, $uid, "Checking for bans matching \002%s\002 on \002%s\002 (\002%d\002) ...",
					$target{info}, ($argv[1] ? join(', ', @{$target{chans}}) : 'all channels'), scalar(@{$target{chans}}));

				for my $chan (@{$target{chans}}) {
					for (sort {$self->{chan}{$self->{lc}{lc($chan)}}{ban}{$b}{ts} <=> $self->{chan}{$self->{lc}{lc($chan)}}{ban}{$a}{ts}} keys(%{$self->{chan}{$self->{lc}{lc($chan)}}{ban}})) {
						next if /^[!@*?~.]+$/;
						$target{banmask} = quotemeta;
						$target{banmask} =~ s/\\\*/.*/g;
						$target{banmask} =~ s/\\\?/./g;

						if ("$target{nick}!$target{ident}\@$target{host}" =~ /^$target{banmask}$/i
							|| ($target{ip} && "$target{nick}!$target{ident}\@$target{ip}" =~ /^$target{banmask}$/i)
							|| ($target{cloak} && "$target{nick}!$target{ident}\@$target{cloak}" =~ /^$target{banmask}$/i)
						) {
							$target{count}++;
							acid::notice($target, $uid, "\002%s\002 on \002%s\002 (set by %s %s ago)",
								$_, $chan, (split(/!/, $self->{chan}{$chan}{ban}{$_}{setby}))[0], acid::duration(time - $self->{chan}{$chan}{ban}{$_}{ts}));
						}
					}
				}

				if ($target{count}) {
					acid::notice($target, $uid, "Listed \002%d\002 bans matching \002%s\002.", $target{count}, $target{nick});
				}
				else {
					acid::notice($target, $uid, "User \002%s\002 doesn't appear to be banned from any of the \002%d\002 channels.",
						$target{nick}, scalar(@{$target{chans}}));
				}
			}

			# MODE
			elsif ($cmd eq 'MODE') {
				if (!$argv[1] || substr($argv[0], 0, 1) ne '#') {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my $chan = $self->{lc}{lc($argv[0])};

				if (!$self->{chan}{$chan}) {
					acid::notice($target, $uid, "Channel \002$argv[0]\002 does not exist.");
					return;
				}

				acid::mode($target, $chan, "@argv[1..$#argv]");
				acid::notice($target, $uid, "Mode \002%s\002 has been set on channel \002%s\002.", "@argv[1..$#argv]", $argv[0]);
				acid::operwall($target, "\002%s\002 used \002%s %s\002 on \002%s\002.", $nick, $cmd, $argv[1], $argv[0]) if !$self->{conf}{abuse};
			}

			# KICK
			elsif ($cmd eq 'KICK') {
				if (!$argv[1] || substr($argv[0], 0, 1) ne '#') {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my $chan = $self->{lc}{lc($argv[0])};
				my $victim = acid::getid($argv[1]);

				if (!$self->{chan}{$chan}) {
					acid::notice($target, $uid, "Channel \002%s\002 does not exist.", $argv[0]);
					return;
				}

				if (!$victim) {
					acid::notice($target, $uid, "There is no user \002%s\002 connected.", $argv[0]);
					return;
				}

				if (!defined($self->{chan}{$chan}{users}{$victim})) {
					acid::notice($target, $uid, "User \002%s\002 is not on channel \002%s\002.", $argv[1], $argv[0]);
					return;
				}
				#if ($self->{client}{$victim}{umode}{N} && $victim ne $uid) {
				#	acid::notice($target, $uid, "Command \002%s\002 may not be used on network administrators.", $cmd);
				#	return;
				#}

				my $reason = ($argv[2] ? "@argv[2..$#argv]" : 'Requested');
				acid::kick($target, $chan, $victim, $reason);
				acid::notice($target, $uid, "User \002%s\002 has been kicked from channel \002%s\002.", $argv[1], $argv[0]);
				acid::operwall($target, "\002%s\002 used \002%s\002 on \002%s\002 on \002%s\002.", $nick, $cmd, $argv[1], $argv[0]) if !$self->{conf}{abuse};
			}

			# SVSMODE
			elsif ($cmd eq 'SVSMODE') {
				if (!$argv[1]) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my $victim = acid::getid($argv[0]);

				if (!$victim) {
					acid::notice($target, $uid, "There is no user \002%s\002 connected.", $argv[0]);
					return;
				}

				#if ($self->{client}{$victim}{umode}{N} && $victim ne $uid) {
				#	acid::notice($target, $uid, "Command \002%s\002 may not be used on other network administrators.", $cmd);
				#	return;
				#}

				acid::svsmode($argv[0], $argv[1]);
				acid::on_mode($self->{conf}{sid}, $victim, $argv[1]);
				acid::notice($target, $uid, "Mode \002%s\002 has been set on user \002%s\002.", $argv[1], $argv[0]);
				acid::operwall($target, "\002%s\002 used \002%s\002 on \002%s\002.", $nick, $cmd, $argv[0]) if !$self->{conf}{abuse};
			}

			# SVSNICK
			elsif ($cmd eq 'SVSNICK') {
				if (!$argv[1] || $argv[1] !~ /^[][}{^`\\|_A-Z\xA0][][}{^`\\|_A-Z\xA0\d-]{0,29}$/i) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my $victim = acid::getid($argv[0]);

				if (!$victim) {
					acid::notice($target, $uid, "There is no user \002%s\002 connected.", $argv[0]);
					return;
				}

				if ($self->{lc}{lc($argv[1])}) {
					acid::notice($target, $uid, "Nick \002%s\002 is already in use.", $argv[1]);
					return;
				}

				#if ($self->{client}{$victim}{umode}{N} && $victim ne $uid) {
				#	acid::notice($target, $uid, "Command \002%s\002 may not be used on other network administrators.", $cmd);
				#	return;
				#}

				acid::svsnick($argv[0], $argv[1]);
				acid::notice($target, $uid, "Nick of user \002%s\002 has been changed to \002%s\002.", $argv[0], $argv[1]);
				acid::operwall($target, "\002%s\002 used \002%s\002 on \002%s\002.", $nick, $cmd, $argv[0]) if !$self->{conf}{abuse};
			}

			# SVSJOIN
			elsif ($cmd eq 'SVSJOIN') {
				if (!$argv[1] || substr($argv[1], 0, 1) ne '#') {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my $victim = acid::getid($argv[0]);
				my $chan = ($self->{chan}{$self->{lc}{lc($argv[1])}} ? $self->{lc}{lc($argv[1])} : $argv[1]);

				if (!$victim) {
					acid::notice($target, $uid, "There is no user \002%s\002 connected.", $argv[0]);
					return;
				}

				if ($self->{chan}{$chan} && defined($self->{chan}{$chan}{users}{$victim})) {
					acid::notice($target, $uid, "User \002%s\002 is on channel \002%s\002 already.", $argv[0], $argv[1]);
					return;
				}

				#if ($self->{client}{$victim}{umode}{N} && $victim ne $uid) {
				#	acid::notice($target, $uid, "Command \002%s\002 may not be used on other network administrators.", $cmd);
				#	return;
				#}

				acid::svsjoin($argv[0], $argv[1]);
				acid::notice($target, $uid, "User \002%s\002 has been joined into \002%s\002.", $argv[0], $argv[1]);
				acid::operwall($target, "\002%s\002 used \002%s\002 on \002%s\002.", $nick, $cmd, $argv[0]) if !$self->{conf}{abuse};
			}

			# SVSPART
			elsif ($cmd eq 'SVSPART') {
				if (!$argv[1] || substr($argv[1], 0, 1) ne '#') {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my $victim = acid::getid($argv[0]);
				my $chan = $self->{lc}{lc($argv[1])};

				if (!$victim) {
					acid::notice($target, $uid, "There is no user \002%s\002 connected.", $argv[0]);
					return;
				}

				if (!$chan) {
					acid::notice($target, $uid, "Channel \002%s\002 does not exist.", $argv[1]);
					return;
				}

				if (!defined($self->{chan}{$chan}{users}{$victim})) {
					acid::notice($target, $uid, "User \002%s\002 is not on \002%s\002.", $argv[0], $argv[1]);
					return;
				}

				#if ($self->{client}{$victim}{umode}{N} && $victim ne $uid) {
				#	acid::notice($target, $uid, "Command \002%s\002 may not be used on other network administrators.", $cmd);
				#	return;
				#}

				acid::svspart($argv[0], $argv[1]);
				acid::notice($target, $uid, "User \002%s\002 has been parted from \002%s\002.", $argv[0], $argv[1]);
				acid::operwall($target, "\002%s\002 used \002%s\002 on \002%s\002.", $nick, $cmd, $argv[0]) if !$self->{conf}{abuse};
			}

			# CAPTURE/UNCAPTURE
			elsif ($cmd eq 'CAPTURE' || $cmd eq 'UNCAPTURE') {
				if (!@argv) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				my $victim = acid::getid($argv[0]);

				if (!$victim) {
					acid::notice($target, $uid, "There is no user \002%s\002 connected.", $argv[0]);
					return;
				}

				if ($self->{client}{$victim}{umode}{o}) {
					acid::notice($target, $uid, "Everyone knows that \002%s\002 cannot be used on IRC operators.", $cmd);
					return;
				}

				if ($cmd eq 'CAPTURE') {
					acid::capture($victim);
					acid::notice($target, $uid, "User \002%s\002 has been captured.", $argv[0]);
				}
				else {
					acid::uncapture($victim);
					acid::notice($target, $uid, "User \002%s\002 has been uncaptured.", $argv[0]);
				}

				acid::operwall($target, "\002%s\002 used \002%s\002 on \002%s\002.", $nick, $cmd, $argv[0]) if !$self->{conf}{abuse};
			}

			# AUTHFLAGS
			elsif ($cmd eq 'AUTHFLAGS') {
				if (!$argv[1]) {
					acid::notice($target, $uid, 'Syntax: %s', $help{$cmd}{syntax});
					acid::notice($target, $uid, "For more information see \002/msg %s HELP %s\002.", $targetnick, $cmd);
					return;
				}

				if ($argv[1] !~ /^[\$=^_><|]+$/) {
					acid::notice($target, $uid, "\002%s\002 is not a valid set of authflags, see \002/msg %s HELP %s\002.", $argv[1], $targetnick, $cmd);
					return;
				}

				my $victim = acid::getid($argv[0]);

				if (!$victim) {
					acid::notice($target, $uid, "There is no user \002%s\002 connected.", $argv[0]);
					return;
				}

				acid::authflags($argv[0], $argv[1]);
				acid::notice($target, $uid, "Authflags \002%s\002 have been added to user \002%s\002.", $argv[1], $argv[0]);
				acid::operwall($target, "\002%s\002 used \002%s\002 on \002%s\002.", $nick, $cmd, $argv[0]) if !$self->{conf}{abuse};
			}

		}
	}
	elsif ((lc($target) eq '#opers' || lc($target) eq '#r') && substr($msg, 0, 1) eq '.') {
		my $myuid = acid::getid($clients{RootServ}{nick});
		return if !defined($self->{client}{$myuid}{chans}{$self->{lc}{lc($target)}});

		my @argv = split(' ', $msg);
		my $cmd = uc(substr(shift(@argv), 1));

		if ($cmd eq 'CLINE') {
			if (!@argv) {
				acid::privmsg($myuid, $target, '.cline server');
				return;
			}

			unless (open CLINES, '<', 'mod/clines.txt') {
				acid::privmsg($myuid, $target, '%s :(', $!);
				return;
			}

			my $count;

			for (<CLINES>) {
				chomp;
				my @cline = split(' ', $_, 4);

				next if $cline[0] !~ /\Q$argv[0]\E/i;

				acid::privmsg($myuid, $target, "/connect %s 0 %s%s\n",
					$cline[0], $cline[1], ($cline[2] ? " (backup: /connect $cline[0] 0 $cline[2])" : ''));
				$count++;
			}

			acid::privmsg($myuid, $target, '%s?!', $argv[0]) if !$count;

			close CLINES;
		}
	}
}

1
